__all__ = ["configure", "shutdown", "client", "events"]

import atexit
from typing import Optional

from . import client, events, locks, session, utils


def configure(api_key: Optional[str] = None,
              rate_limit: Optional[int] = None,
              enable_events: bool = False,
              **session_kwargs) -> client.ApiClient:
    """
    Configure this fredio application

    :param api_key: FRED API key. If None, will be retrived from environment
    variable FRED_API_KEY
    :param rate_limit: API rate limit. Must be <= 120.
    :param enable_events: Should fredio events be enabled?
    :param session_kwargs: Keyword arguments passed to aiohttp.ClientSession
    """

    api_key = api_key or client.get_api_key()

    if api_key is None:
        msg = "Api key must be provided or passed as environment variable FRED_API_KEY"
        raise ValueError(msg)

    ses = session.Session(**session_kwargs)

    client_ = client.get_client()
    client_.set_session(ses)
    client_.set_defaults(api_key=api_key, file_type="json")

    if rate_limit is not None:
        locks.set_rate_limit(rate_limit)

    if enable_events:
        events.listen()

    return client_


def shutdown():
    """
    Shutdown this fredio appilication. Cancels all running tasks in this event
    loop and closes the aiothttp.ClientSession
    """
    # Flush all events and cancel
    utils.loop.run_until_complete(events.cancel())

    # Cancel other tasks (ratelimiter & others)
    utils.cancel_running_tasks()

    # Close aiohttp session
    client.get_client().close_session()


atexit.register(shutdown)



__all__ = ["Session"]

import asyncio
import itertools
import logging
from functools import partial
from typing import Any, List, Dict, Optional

import jsonpath_rw
from aiohttp import ClientSession, ClientResponseError
from yarl import URL

from . import events
from . import locks


logger = logging.getLogger(__name__)


class Session(object):
    """
    Async request/response handling with some api-specific logic
    """

    def __init__(self, **kwargs):

        self._session_cls = ClientSession
        self._session_kws = kwargs

        self._session = None

    @property
    def session(self) -> ClientSession:
        """
        Return a ClientSession instance (cached)
        """
        if self._session is None:
            logger.info("Initializing %s" % self._session_cls.__name__)
            self._session = self._session_cls(**self._session_kws)
        return self._session

    async def request(self,
                      method: str,
                      url: URL,
                      retries: int = 0,
                      **kwargs) -> Dict[Any, Any]:
        """
        Wraps ClientSession.request() with rate limiting and handles retry logic

        :param method: Request method
        :param url: URL
        :param retries: Maximum number of request retries
        :param kwargs: Request parameters
        """
        kwargs.setdefault("raise_for_status", True)
        ratelimiter = locks.get_rate_limiter()

        attempts = 0
        while attempts <= retries:
            try:
                async with ratelimiter:
                    async with self.session.request(method, url, **kwargs) as response:

                        # Emit an event with name corresponding to the final endpoint
                        if events.running():
                            name = response.url.path.split("/")[-1]
                            await events.produce(name, response)

                        return await response.json()

            except ClientResponseError as e:
                attempts += 1
                logging.error(e)

                if attempts > retries:
                    raise
                elif e.status == 429:
                    backoff = ratelimiter.get_backoff()
                    logger.debug("Retrying request in %d seconds" % backoff)
                    await asyncio.sleep(backoff)
                else:
                    raise

    async def get(self,
                  url: URL,
                  jsonpath: Optional[str] = None,
                  retries: int = 3,
                  **parameters) -> List[Dict]:
        """
        Will await a single request to get the first batch of data before executing
        subsequent requests (if required) according to offset logic. Jsonpath query
        is optionally executed on json response data from each request.

        :param url: yarl URL
        :param jsonpath: Optional jsonpath query to process response data
        :param retries: Retry count, passed to Session.request
        :param parameters: API Request parameters
        """

        if parameters:
            url = url.update_query(**parameters)

        getter = partial(self.request, "GET", retries=retries)

        response = [await getter(url)]

        count = response[0].get("count")
        limit = response[0].get("limit")
        offset = response[0].get("offset")

        logger.debug("Count: %s, Limit: %s, Offset: %s" % (count, limit, offset))

        if any((count, limit, offset)):

            coros = list(map(
                lambda x: getter(url=url.update_query(offset=x)),
                range(limit + offset, count, limit)
            ))

            logger.debug("Planning %s additional requests" % len(coros))
            response.extend(await asyncio.gather(*coros))

        if jsonpath:
            parsed = jsonpath_rw.parse(jsonpath)
            mapped = map(lambda x: [i.value for i in parsed.find(x)], response)
            return list(itertools.chain.from_iterable(mapped))

        return response

    async def close(self) -> None:
        """
        Close the ClientSession
        """
        await self.session.close()
        self._session = None



FRED_API_ENDPOINTS = (
    "category",
    "category/children",
    "category/related",
    "category/series",
    "category/tags",
    "category/related_tags",

    "releases",
    "releases/dates",

    "release",
    "release/dates",
    "release/series",
    "release/sources",
    "release/tags",
    "release/related_tags",
    "release/tables",

    "series",
    "series/categories",
    "series/observations",
    "series/release",
    "series/search",
    "series/search/tags",
    "series/search/related_tags",
    "series/tags",
    "series/updates",
    "series/vintagedates",

    "sources",

    "source",
    "source/releases",

    "tags",
    "tags/series",
    "related_tags"
)

FRED_API_URL = "https://api.stlouisfed.org/fred"
FRED_DOC_URL = "https://fred.stlouisfed.org/docs/api/fred"

FRED_API_RATE_LIMIT = 120
FRED_API_RATE_RESET = 60
FRED_API_FILE_TYPE = "json"  # other XML option but we dont want that



__all__ = ["get_rate_limiter", "set_rate_limit"]

import asyncio
import logging
import math
import time
from collections import deque

from . import const
from . import utils

logger = logging.getLogger(__name__)


class RateLimiter(asyncio.BoundedSemaphore):
    """
    Rate-limiting implementation using a BoundedSemaphore.
    Locks are only released on a specified interval via the `replenish` loop
    which runs as a background task.
    """

    _bound_value: int
    _loop: asyncio.BaseEventLoop

    def __init__(self,
                 limit: int = const.FRED_API_RATE_LIMIT,
                 period: int = const.FRED_API_RATE_RESET,
                 *,
                 loop: asyncio.BaseEventLoop = utils.loop):

        super(RateLimiter, self).__init__(limit, loop=loop)

        # Store the delta between monotonic and system time to define
        # a point of reference
        self._deltat = time.time() - self._loop.time()

        self._period = period
        self._releases = deque()
        self._task = None

    def _timer(self) -> float:
        """
        Monotonic time synced with the system
        """
        return self._loop.time() + self._deltat

    @property
    def started(self) -> bool:
        return self._task is not None

    def get_backoff(self, ceil: bool = True) -> float:
        """
        Get number of seconds until the next count

        :param ceil: Round backoff time up to the closest integer
        """
        backoff = self._period - self._timer() % self._period

        if ceil:
            return float(math.ceil(backoff))
        return backoff

    def get_counter(self) -> int:
        """
        Get number of periods since timer start
        """
        return int(self._timer() // self._period)

    def start(self) -> bool:
        """
        Create background replenishment task. Can be called more than once.
        """
        if not self.started:
            self._task = self._loop.create_task(self.replenish())
        return True

    def stop(self) -> bool:
        """
        Cancel the replenishment task
        """
        if self.started:
            self._task.cancel()
            self._task = None
        return True

    async def acquire(self) -> bool:
        """
        Acquire a lock
        """
        self.start()

        return await super(RateLimiter, self).acquire()

    async def replenish(self) -> None:
        """
        Run a continuous loop to periodically release all locks from the
        previous count.

        https://stackoverflow.com/a/48685838
        """

        while True:
            while self._releases:

                # Peek and compare counters
                # A lock will only be released if scheduled from a previous count
                if self._releases[0][1] <= self.get_counter():

                    ts, ct = self._releases.popleft()
                    super(RateLimiter, self).release()

                    logger.debug("Released lock from counter %d (elapsed: %.4f)"
                                 % (ct, self._timer() - ts))
                else:
                    break

            # Sleep until the next count
            await asyncio.sleep(self.get_backoff(False))

    def release(self) -> None:
        """
        Schedule a lock to be released in the next counter
        """
        self._releases.append((self._timer(), self.get_counter()))


_ratelimiter = RateLimiter()


def get_rate_limiter() -> RateLimiter:
    """
    Get the global ratelimiter
    """
    return _ratelimiter


def set_rate_limit(limit: int = const.FRED_API_RATE_LIMIT) -> bool:
    """
    Reset the global ratelimiter with a new limit.

    :param limit: Number of requests per minute. Should be < 120
    """
    if limit > const.FRED_API_RATE_LIMIT:
        raise ValueError("Limit must be <= %d" % const.FRED_API_RATE_LIMIT)

    global _ratelimiter

    _ratelimiter.stop()
    _ratelimiter = RateLimiter(limit)
    _ratelimiter.start()

    return True



__all__ = [
    "Event", "produce", "consume", "listen",
    "running", "register", "on_event"
]

import asyncio
import logging
from typing import Any, Awaitable, Callable, Dict, Optional, Union

from . import utils


logger = logging.getLogger(__name__)

queue = asyncio.Queue()

_events: Dict[str, "Event"] = dict()  # TODO: This may need a lock
_task: Optional[asyncio.Task] = None


class Event(object):

    def __init__(self, name: str):
        self.name = name

        self._handlers = []
        self._frozen = False

    def add(self, handler: Callable) -> None:
        """
        Add a handler to this event. Sync functions will be wrapped as
        a coroutine.

        :param handler: Callable function or coroutine
        """
        if self._frozen:
            raise RuntimeError("Cannot modify frozen event")

        self._handlers.append(asyncio.coroutine(handler))
        logger.info("Registered handler '%s' for event '%s'"
                    % (handler.__name__, self.name))

    def freeze(self):
        self._frozen = True

    async def apply(self, *args, **kwargs) -> None:
        """
        Call all event handlers for this event.
        """
        logger.debug("Received event '%s'" % self.name)

        for handler in self._handlers:
            await handler(*args, **kwargs)


async def produce(name: str, data: Any, q: asyncio.Queue = queue) -> None:
    """
    Places a tuple of (name, data) on the queue.

    :param name: name
    :param data: data
    :param q: queue
    """
    return await q.put((name, data))


async def consume(q: asyncio.Queue = queue) -> None:
    """
    Consume an event from the queue, and create a task to call all
    handlers for this event.

    :param q: queue
    """
    name, event = await q.get()
    if name in _events:
        utils.loop.create_task(_events[name].apply(event))
    q.task_done()


def flush(timeout: Optional[Union[float, int]] = None) -> Awaitable:
    """
    Flush all remaining tasks from the queue

    :param timeout: Timeout (seconds) after which TimeoutError will be raised.
    """
    return asyncio.wait_for(queue.join(), timeout)


async def cancel(timeout: Optional[Union[float, int]] = None) -> None:
    """
    Cancel the consumer task. Pending events in the queue will be flushed
    with a timeout.

    :param timeout: Timeout (seconds) after which TimeoutError will be raised.
    """
    global _task

    if running():
        logger.debug("Flushing %d remaining tasks" % queue.qsize())

        try:
            await flush(timeout)
        except asyncio.TimeoutError as e:
            logger.exception(e)
        finally:
            _task.cancel()
            _task = None


def listen() -> bool:
    """
    Start a background task to consume events
    """
    global _task

    async def _listen() -> None:
        while True:
            await consume(queue)

    if not running():

        for ev in _events.values():
            ev.freeze()

        logger.info("Listening for events: \n%s" % "\n".join(_events.keys()))
        _task = utils.loop.create_task(_listen())

    return True


def running() -> bool:
    """
    Is the consumer task running?
    """
    if _task is not None:
        return not _task.done()
    return False


def register(name: str, fn: Callable[..., Awaitable]) -> None:
    """
    Register an event handler

    :param name: Name of the event handler
    :param fn:
    """
    _events.setdefault(name, Event(name)).add(fn)


# --- decorators --- #

def on_event(name: str) -> Callable:
    """
    Register an event handler

    :param name: Name of the event that this handler should process
    """
    def wrapper(fn):
        register(name, fn)
        return fn
    return wrapper



__all__ = ["get_all_tasks"]

import asyncio
import logging
import sys
from typing import Set

logger = logging.getLogger(__name__)

# Main event loop
loop = asyncio.get_event_loop()


def get_all_tasks() -> Set[asyncio.Task]:
    """
    Get all tasks from main event loop
    """
    if sys.version_info < (3, 7, 0):
        return asyncio.Task.all_tasks(loop=loop)
    else:
        return asyncio.all_tasks(loop=loop)


def cancel_running_tasks() -> None:
    """
    Cancel all running tasks in this loop
    """
    for task in get_all_tasks():
        if not task.done():
            logger.info("Cancelling task %s" % task)
            task.cancel()



__all__ = ["ApiClient", "get_api_key", "get_client", "get_endpoints"]

import asyncio
import logging
import os
import urllib
import webbrowser
from typing import Any, Dict, List, Optional, Type

from aiohttp.typedefs import StrOrURL
from pandas import DataFrame
from yarl import URL

from . import const
from . import utils
from . import session

logger = logging.getLogger(__name__)


class ApiClient(object):
    """
    Structure containing a top level URL and child endpoints
    """

    _children: Dict[str, "ApiClient"]  # TODO: get rid of mutability
    _defaults: Dict[Any, Any] = dict()  # Default parameters for all client instances
    _session: session.Session = None

    def __init__(self, url: StrOrURL):
        super(ApiClient, self).__init__()

        self._children = dict()
        self._url = URL(url, encoded=True)

    def __getattribute__(self, item: Any) -> Any:
        """
        Hijack to allow for accessing children using dot notation
        """
        try:
            return super(ApiClient, self).__getattribute__(item)
        except AttributeError:
            children = super(ApiClient, self).__getattribute__("_children")
            if item not in children.keys():
                raise
            return children[item]

    def __repr__(self):  # pragma: no-cover
        return f'{self.__class__.__name__}<{self._url}>'

    def _encode_url(self) -> URL:
        """
        Create new URL with default query parameters
        """
        # safe_chars is hard-coded as these chars are used for tag requests etc
        query = urllib.parse.urlencode(self._defaults, safe=",;")
        return self._url.with_query(query)

    @classmethod
    def set_defaults(cls, **params) -> Type["ApiClient"]:
        """
        Set default query parameters for all endpoints

        :param params: Request parameters that will be passed for every request
        for all ApiClient instances
        """
        cls._defaults = params
        return cls

    @classmethod
    def set_session(cls, ses: session.Session) -> Type["ApiClient"]:
        """
        Set the Session for this class

        :param ses: Session used by all ApiClient instances
        """
        cls._session = ses
        return cls

    @classmethod
    def close_session(cls):
        """
        Close the client session
        """
        if cls._session is not None:
            logger.info("Closing client session")
            return utils.loop.run_until_complete(cls._session.close())

    @property
    def children(self) -> Dict[str, "ApiClient"]:
        """
        Get client children
        """
        return self._children

    @property
    def docs(self) -> "_ApiDocs":
        return _ApiDocs(self._url)

    @property
    def url(self) -> URL:
        """
        Encode this client's URL with query parameters
        """
        return self._encode_url()

    def aget(self, **kwargs) -> asyncio.Task:
        """
        Run an awaitable Task to get results from this endpoint

        :param kwargs: Keyword arguments passed to Session.get
        """
        coro = self._session.get(self.url, **kwargs)
        task = utils.loop.create_task(coro)
        return task

    def get(self, **kwargs) -> List[Dict]:
        """
        Get request results as a list of JSON. This method is blocking.

        :param kwargs: Keyword arguments passed to Session.get
        """
        return utils.loop.run_until_complete(self.aget(**kwargs))

    def get_pandas(self, **kwargs) -> DataFrame:
        """
        Get request results as a pandas DataFrame. This method is blocking.

        :param kwargs: Keyword arguments passed to Session.get
        """
        return DataFrame.from_records(self.get(**kwargs))


class _ApiDocs:
    """
    Helper class containing webbrowser.open methods to open FRED documentation
    corresponding to an ApiClient URL
    """

    def __init__(self, url):
        self.url = url

    def make_url(self) -> URL:
        subpath = self.url.path.replace("/fred", "").lstrip("/").replace("/", "_")

        if subpath:
            subpath += ".html"

        return URL(const.FRED_DOC_URL) / subpath

    def open(self) -> bool:
        return webbrowser.open(str(self.make_url()))

    def open_new(self) -> bool:
        return webbrowser.open_new(str(self.make_url()))

    def open_new_tab(self) -> bool:
        return webbrowser.open_new_tab(str(self.make_url()))

    open.__doc__ = webbrowser.open.__doc__
    open_new.__doc__ = webbrowser.open_new.__doc__
    open_new_tab.__doc__ = webbrowser.open_new_tab.__doc__


def get_api_key() -> Optional[str]:
    """
    Get API key from FRED_API_KEY environment variable
    """
    return os.environ.get("FRED_API_KEY", None)


def get_client() -> ApiClient:
    """
    Construct an ApiClient and add FRED endpoints.
    """

    client = ApiClient(const.FRED_API_URL)
    add_endpoints(client, *const.FRED_API_ENDPOINTS)

    return client


def add_endpoints(client: ApiClient, *endpoints) -> None:
    """
    Add an endpoint to the client instance

    :param client: ApiClient
    :param endpoints: URL subpaths to add to the ApiClient structure
    """
    for ep in endpoints:
        parent, *child = ep.split("/", 1)
        newpath = client.children.setdefault(parent, ApiClient(client.url / parent))
        if len(child):
            add_endpoints(newpath, child[0])


def get_endpoints(client: ApiClient) -> List[URL]:
    """
    Get all registered URLs from an ApiClient

    :param client: ApiClient
    """
    endpoints = []
    for node in client.children.values():
        if isinstance(node, ApiClient):
            endpoints.extend(get_endpoints(node))
    endpoints.append(client.url)
    return endpoints



import os
from setuptools import find_packages, setup


dirname = os.path.dirname(__file__)

with open(os.path.join(dirname, "README.md"), "r") as f:
    long_description = f.read()

setup(
    name="fredio",
    version="0.1.0a1",
    description="Asynchronous python client for the FRED® API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    url="https://github.com/bgrams/fredio",
    author="Brandon Grams",
    license="MIT",
    python_requires=">=3.6",
    packages=find_packages(include=["fredio", "fredio.*"]),
    install_requires=[
        "aiohttp>=3.0,<4.0",
        "jsonpath-rw",
        "pandas",
        "yarl>=1.0,<2.0"
    ],
    test_suite="tests"
)
